/*
 * cfgtool - A Configurator tool for ROSAND Technologies products.
 *
 * Inspired from AVRDUDE program.
 *
 * Copyright (C) 2000-2005  Brian S. Dean <bsd@bsdhome.com>
 * Copyright 2007-2009 Joerg Wunsch <j@uriah.heep.sax.de>
 * Copyright (c) 2011, Rosand Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#include "8051.h"

#include "pgm.h"
#include "lists.h"
#include "update.h"

/*
 * Read the the specified configuration into the
 * corresponding buffer of the part pointed to by 'p'.  If size =
 * 0, read the entire contents, otherwise, read 'size' bytes.
 *
 * Return the number of bytes read, or < 0 if an error occurs.
 */
int part_read(PROGRAMMER * pgm, PART * p, char * memtype, int size,
             int index)
{
  MEM * mem;
  int rc;

  mem = locate_mem(p, memtype);
  if (mem == NULL) {
    fprintf(stderr, "No \"%s\" configuration for product %s\n",
            memtype, p->desc);
    return -1;
  }

  if (size == 0) {
    size = mem->size;
  }

  if(mem->pollindex == 0x41) {
	  rc = pgm->paged_load(pgm, p, mem, index, size);
  } else {
	  rc = pgm->paged_load(pgm, p, mem, 0, size);
  }
  return rc;
}


/*
 * Write the whole memory region of the specified memory from the
 * corresponding buffer of the part pointed to by 'p'.  Write up to
 * 'size' bytes from the buffer.  Data is only written if the new data
 * value is different from the existing data value.  Data beyond
 * 'size' bytes is not affected.
 *
 * Return the number of bytes written, or -1 if an error occurs.
 */
int part_write(PROGRAMMER * pgm, PART * p, char * memtype, int size,
              int index)
{
  int   rc;
  MEM * mem;

  mem = locate_mem(p, memtype);
  if (mem == NULL) {
    fprintf(stderr, "No \"%s\" parameter for product %s\n",
            memtype, p->desc);
    return -1;
  }

  pgm->err_led(pgm, OFF);

  if (size == 0) {
    size = mem->size;
  }

  rc = pgm->paged_write(pgm, p, mem, index, size);

  if(rc < 0) {
	/*
     * make sure the error led stay on if there was a previous write
     * error, otherwise it gets cleared in part_write_byte()
     */
	pgm->err_led(pgm, ON);
  }

  return size;
}


/*
 * Verify the memory buffer of p with that of v.  The byte range of v,
 * may be a subset of p.  The byte range of p should cover the whole
 * chip's memory size.
 *
 * Return the number of bytes verified, or -1 if they don't match.
 */
int part_verify(PART * p, PART * v, char * memtype, int size)
{
  int i;
  unsigned char * buf1, * buf2;
  int vsize;
  MEM * a, * b;

  a = locate_mem(p, memtype);
  if (a == NULL) {
    fprintf(stderr,
            "part_verify(): parameter \"%s\" not defined for product %s\n",
            memtype, p->desc);
    return -1;
  }

  b = locate_mem(v, memtype);
  if (b == NULL) {
    fprintf(stderr,
            "part_verify(): parameter \"%s\" not defined for product %s\n",
            memtype, v->desc);
    return -1;
  }

  buf1  = a->buf;
  buf2  = b->buf;
  vsize = a->size;

  if (vsize < size) {
    fprintf(stderr,
            "%s: WARNING: requested verification for %d bytes\n"
            "%s%s parameter field only contains %d bytes\n"
            "%sOnly %d bytes will be verified.\n",
            progname, size,
            progbuf, memtype, vsize,
            progbuf, vsize);
    size = vsize;
  }

  for (i=0; i<size; i++) {
    if (buf1[i] != buf2[i]) {
      fprintf(stderr,
              "%s: verification error, first mismatch at byte 0x%04x\n"
              "%s0x%02x != 0x%02x\n",
              progname, i,
              progbuf, buf1[i], buf2[i]);
      return -1;
    }
  }

  return size;
}
